/******************************************************************************
 * 
 * 
 * 
 ******************************************************************************
 */
#ifndef lint
static char *RCSid = "$Header: /home/beason/ksmith/moldy/RCS/writeIO.c,v 1.5 2002/12/15 03:23:30 beason Exp beason $";
#endif
/*========================== Library include files ===========================*/
#include	"defs.h"
/*========================== Library include files ===========================*/
#include	<math.h>
#include	"string.h"
#include	"stddef.h"
#include	<stdio.h>
/*========================== program include files ===========================*/
#include "structs.h"
#include "messages.h"
/*========================== External function declarations ==================*/
extern void     update_visualization(float **positions, int rays, float **raypos,
				     float h);
gptr            *talloc();	       /* Interface to memory allocator       */
void            tfree();	       /* Free allocated memory	      	      */
void            afree();	       /* Free allocated array	      	      */
void            step_1();	       /* Step co-ordinates by Beeman algrthm */
void            step_2();	       /* Step velocities at above            */
void            beeman_2();	       /* As above for individual components  */
void            make_sites();	       /* Construct site coordinate arrays    */
void            mol_force();	       /* Calculare molecular from site force */
void            mol_torque();	       /* Calculate torques from site forces  */
void            rotate();	       /* Perform rotations given quaternions */
void            newton();	       /* Calculate accelerations from forces */
void            euler();	       /* Get quat 2nd derivatives            */
void            parinello();	       /* Get correction to c of m accns      */
void            rahman();	       /* Get h 2nd derivatives	              */
void		wentzcovitch();	       /* Cell accn using Wentzcovitch's H    */
void            energy_dyad();	       /* Calculate mvv for stress term       */
void            force_calc();	       /* Calculate direct-space forces       */
double          dist_pot();	       /* Returns integrated potential fn     */
void            ewald();	       /* Get Ewald sum forces		      */
void            dump();		       /* Maintain and write dump data files  */
void            zero_real();	       /* Clear area of memory		      */
void            zero_double();	       /* Clear area of memory		      */
void            invert();	       /* Matrix inverter		      */
double          det();		       /* Returns matrix determinant	      */
void            mat_vec_mul();	       /* 3 x 3 Matrix by Vector multiplier   */
void            mean_square();	       /* Caluculates mean square of args     */
void            rdf_calc();	       /* Accumulate and bin rdf	      */
double 		value();	       /* Return thermodynamic average	      */
double 		roll_av();	       /* Return thermodynamic average	      */
double          vdot();		       /* Fast vector dot product	      */
double		sum();		       /* Fast vector sum.		      */
void            vscale();	       /* Vector by constant multiply	      */
double          vec_dist();	       /* normalised vector distance	      */
void		thermalise();	       /* Randomize velocities to given temp  */
double		trans_ke();	       /* Compute translational kinetic en.   */
double		rot_ke();	       /* Compute rotational kinetic en.      */
void            hoover_tr();           /* Corrects forces due to thermostat   */
void            hoover_rot();          /* Corrects forces due to thermostat   */
double          gaussiant();           /* Return Force*vel                    */
double          gaussianr1();          /* Return Torque*omega                 */
double          gaussianr2();          /* Return omega*I*omega                */
void            q_conj_mul();          /* Quat. conjugated x by quat. dot     */
void	inhibit_vectorization();       /* Self-explanatory dummy              */
void            kernel();              /* Potential function evaluation       */
#ifdef SPMD
void		par_rsum();
void		par_dsum();
#endif
#ifdef HAVE_STDARG_H
gptr		*arralloc(size_mt,int,...); /* Array allocator		      */
void		note(char *, ...);	/* Write a message to the output file */
void		message(int *, ...);	/* Write a warning or error message   */
#else
gptr		*arralloc();	        /* Array allocator		      */
void		note();			/* Write a message to the output file */
void		message();		/* Write a warning or error message   */
#endif
/*========================== External data references ========================*/
extern contr_mt control;                    /* Main simulation control parms. */
extern int 	ithread, nthreads;
/*========================== Macros ==========================================*/
#define ITER_MAX 10
#define	CONVRG	1.0e-7
/*  Can't rely on ANSI yet. */
#ifndef DBL_MIN
#   define DBL_MIN 1.0e-36
#endif
/*========================== Cache Parameters=================================*/
/* The default values are for the Cray T3D but are probably good enough 
 *  for most other systems too. */
#ifndef NCACHE
#   define NCACHE (256*sizeof(double)/sizeof(real))
#endif
#ifndef NLINE
#   define NLINE  (4*sizeof(double)/sizeof(real))
#endif
/*============================================================================*/
void 
write_interactive_output(sys, control, rays, species, site_info)
system_mp       sys;		       /* Pointer to system info	 (in) */
contr_mp        control;               /* Pointer to control info (in)        */
rays_mp         rays;                  /* Pointer to rays_mt ray def (in)     */
spec_mt         species[];	       /* Array of species info	 (in)         */
site_mt         site_info[];	       /* Array of site info structures (in)  */
{
   //FILE         *dumpf;
   double       **site = (double**)arralloc(sizeof(double),2,
                                            0,2,0,sys->nsites-1);
   spec_mt      *spec;
   int          imol, isite, itot=1, ispec=1, iposn;
   int          is;
   char         *fname;
   float        **positions; /* [sys->nmols*3][3] */
   float        **raypos;    /* [rays->num][6] */
   int i, j, plane, planes[3], crossing, axis;
   int first_plane = -1;
   int first_crossing = -1;
   int intersections = 0;
   float t;
   vec_mp       pos, dir;
   vec_mt       x[2];
   float cross[2] = { -(sys->h[0][0]/2.0), (sys->h[0][0]/2.0) };
   float eps = 0.01;
   int ok;

   /* init positions */
   positions = (float**) malloc(sizeof(float*)*sys->nmols*3);
   for (i=0;i<sys->nmols*3;i++)
     positions[i] = (float*) malloc(sizeof(float)*3);

   /* init raypos */
   raypos = (float**) malloc(sizeof(float*)*rays->num);
   for (i=0;i<rays->num;i++)
     raypos[i] = (float*) malloc(sizeof(float)*6);

   /*if( (*dumpf = fopen(fname, "r+b")) == NULL) */
   //if( (dumpf = fopen("h20.output", "w")) == NULL)  /* Open dump file     */
      //message(NULLI, NULLP, WARNING, DOERRR, fname, strerror(errno));

   for(spec = species; spec < species+sys->nspecies; ispec++, spec++)
   {
     make_sites(sys->h, spec->c_of_m, spec->quat, spec->p_f_sites,
                site, spec->nmols, spec->nsites, MOLPBC);

     isite = 0;
     iposn = 0;
     for(imol = 0; imol < spec->nmols; imol++)
     {
       for(is = 0; is < spec->nsites; is++)
       {
         if(fabs(site_info[spec->site_id[is]].mass) != 0) {
            //(void)fprintf(dumpf,"%10.6f %10.6f %10.6f\n", site[0][isite], site[1][isite], site[2][isite]);
            //printf("\n isite %d iposn %d", isite,iposn);
            positions[iposn][0] = site[0][isite];
            positions[iposn][1] = site[1][isite];
            positions[iposn][2] = site[2][isite];
            iposn++;
         }
         isite++;
         itot++;
       }
     }
   }

   /* fill out raypos */
   for (i=0; i<rays->num;i++){
      pos = &(rays->pos[i]);
      dir = &(rays->dir[i]);
/*
      fprintf(stderr, "fill: (%4f, %4, %4f) -> (%4f, %4f, %4f)\n",
	      (*pos)[0], (*pos)[1], (*pos)[2],
	      (*dir)[0], (*dir)[1], (*dir)[2]);
*/
      /* foreach intersection */
      intersections = 0;
      for (j=0; j<2; j++){
         /* foreach plane orientation */
	 for (plane=0; plane<3; plane++)
            /* foreach crossing */
	    for (crossing=0; crossing<2; crossing++){

	       /* skip same-plane intersection for 2nd intersection */
	       if (j==1 && crossing==first_crossing && plane==first_plane)
		  continue;
	       
	       /* solve ray-plane intersection */
	       if ((*dir)[plane]){
  	         t = ( cross[crossing] - (*pos)[plane] ) / (*dir)[plane];
		 planes[0] = plane;
		 planes[1] = (plane+1)%3;
		 planes[2] = (plane+2)%3;
		 /* find intersection */
		 ok = 1;
		 for (axis=0; axis<3; axis++){
		    x[j][axis] = (*pos)[axis] + t*(*dir)[axis];

		    /* check if within bounds */
		    if (x[j][axis] > cross[1]+eps ||
			x[j][axis] < cross[0]-eps )
		       ok = 0;
		 }
		 if (ok != 1)
		    continue;

		 /* assume ok */
		 intersections++;
		 
		 /* set variables */
		 if (j == 0){
		    first_crossing = crossing;
		    first_plane = plane;
		 }

		 goto found_intersection;
	       }
	    }
	found_intersection:
      }
	 
     done_with_crossings:

      /* make sure we have two intersections */
      for (j=intersections;j<2; j++){
	 for (axis=0; axis<3; axis++)
            x[j][axis] = 0.0;
      }

      /* copy data into raypos */
      for (crossing=0; crossing<2; crossing++)
	 for (axis=0; axis<3; axis++)
	    raypos[i][crossing*3+axis] = x[crossing][axis];
      
   }
   
   //fclose(dumpf);
   update_visualization(positions, rays->num, raypos, sys->h[0][0]);
}
